
from enum import IntEnum

class MetadataType(IntEnum):
    Title = 0
    Artist = 1
    Creator = 2
    Version = 3
    Source = 4
    Tags = 5
    VideoDataOffset = 6
    VideoDataLength = 7
    VideoHash = 8
    BeatmapSetID = 9
    Genre = 10
    Language = 11
    TitleUnicode = 12
    ArtistUnicode = 13
    Unknown = 9999
    Difficulty = 10000
    PreviewTime = 10001
    ArtistFullName = 10002
    ArtistTwitter = 10003
    SourceUnicode = 10004
    ArtistURL = 10005
    Revision = 10006
    PackID = 10007




from .utils import uint32_slice_to_byte_slice
from . import crypto
from typing import List
import numpy as np

class SimpleCryptor:
    def __init__(self, key: List[int]) -> None:
        # Pre-compute byte key for better performance
        self.key = np.array(uint32_slice_to_byte_slice(key), dtype=np.uint8)

    def encrypt_bytes(self, buf: bytearray) -> None:
        crypto.encrypt_bytes(buf, self.key)

    def decrypt_bytes(self, buf: bytearray) -> None:
        crypto.decrypt_bytes(buf, self.key)




from osz2.keys import generate_osf2_key, generate_osz2_key
from osz2 import MetadataType
from hashlib import md5

class TestKeyGeneration:
    def test_osz2_key_generation(self) -> None:
        metadata = {
            MetadataType.Creator: "TestCreator",
            MetadataType.BeatmapSetID: "12345",
        }
        key = generate_osz2_key(metadata)
        expected = md5(b"TestCreatoryhxyfjo512345").digest()
        assert key == expected

    def test_osf2_key_generation(self) -> None:
        metadata = {
            MetadataType.Title: "TestTitle",
            MetadataType.Artist: "TestArtist",
        }
        key = generate_osf2_key(metadata)
        expected = md5("\x08TestTitle4390gn8931iTestArtist".encode()).digest()
        assert key == expected




from osz2 import Osz2Package, MetadataType, KeyType
from pytest import fixture
from pathlib import Path

TESTS_DIR = Path(__file__).parent
FILES_DIR = TESTS_DIR / "files"
OSF2_FILES = list(FILES_DIR.glob("*.osf2"))

class TestOsf2Package:
    @fixture(params=OSF2_FILES, ids=lambda p: p.stem)
    def osf2_path(self, request) -> Path:
        return request.param

    def test_from_file(self, osf2_path: Path) -> None:
        package = Osz2Package.from_file(str(osf2_path), key_type=KeyType.OSF2)
        assert package is not None
        assert len(package.files) > 0
        assert len(package.metadata) > 0

    def test_from_bytes(self, osf2_path: Path) -> None:
        data = osf2_path.read_bytes()
        package = Osz2Package.from_bytes(data, key_type=KeyType.OSF2)
        assert package is not None
        assert len(package.files) > 0

    def test_metadata_only(self, osf2_path: Path) -> None:
        package = Osz2Package.from_file(str(osf2_path), metadata_only=True, key_type=KeyType.OSF2)
        assert len(package.metadata) > 0
        assert len(package.files) == 0

    def test_metadata_contains_required_fields(self, osf2_path: Path) -> None:
        package = Osz2Package.from_file(str(osf2_path), metadata_only=True, key_type=KeyType.OSF2)
        # 'Title' and 'Artist' are required for osf2 key generation
        assert MetadataType.Title in package.metadata
        assert MetadataType.Artist in package.metadata




from .simple_cryptor import SimpleCryptor
from . import crypto

from typing import List
import numpy as np

MAX_WORDS = 16
MAX_BYTES = MAX_WORDS * 4

class XXTEA:
    """XXTEA implements the Corrected Block TEA algorithm"""

    def __init__(self, key: List[int]) -> None:
        self.cryptor = SimpleCryptor(key)
        self.key = np.array(key, dtype=np.uint32)

        # Pre-compute all possible key permutations for faster lookup
        self.key_table = np.array([[key[i ^ e] for i in range(4)] for e in range(4)], dtype=np.uint32)
        self.n = 0

    def decrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self.encrypt_decrypt(buffer, start, count, False)

    def encrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self.encrypt_decrypt(buffer, start, count, True)

    def encrypt_decrypt(self, buffer: bytearray, buf_start: int, count: int, encrypt: bool) -> None:
        full_word_count = count // MAX_BYTES
        left_over = count % MAX_BYTES

        # Process full blocks
        if full_word_count > 0:
            if encrypt:
                self.encrypt_full_blocks(buffer, buf_start, full_word_count)
            else:
                self.decrypt_full_blocks(buffer, buf_start, full_word_count)

        if left_over == 0:
            return

        leftover_start = buf_start + full_word_count * MAX_BYTES
        self.n = left_over // 4

        if self.n > 1:
            if encrypt:
                self.encrypt_words(self.n, self.key, buffer, leftover_start)
            else:
                self.decrypt_words(self.n, self.key, buffer, leftover_start)

            left_over -= self.n * 4
            if left_over == 0:
                return

            leftover_start += self.n * 4

        remaining = buffer[leftover_start:leftover_start + left_over]

        if encrypt:
            self.cryptor.encrypt_bytes(remaining)
        else:
            self.cryptor.decrypt_bytes(remaining)

        buffer[leftover_start:leftover_start + left_over] = remaining

    def encrypt_full_blocks(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Use parallel processing for multiple blocks
        if full_word_count >= 4:
            self.encrypt_full_blocks_parallel(buffer, buf_start, full_word_count)
            return

        # Sequential for small data
        for i in range(full_word_count):
            offset = buf_start + i * MAX_BYTES
            self.encrypt_fixed_word_array(self.key, buffer, offset)

    def decrypt_full_blocks(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Use parallel processing for multiple blocks
        if full_word_count >= 4:
            self.decrypt_full_blocks_parallel(buffer, buf_start, full_word_count)
            return

        # Sequential for small data
        for i in range(full_word_count):
            offset = buf_start + i * MAX_BYTES
            self.decrypt_fixed_word_array(self.key, buffer, offset)

    def encrypt_full_blocks_parallel(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Convert buffer slice to numpy array for parallel processing
        buffer_size = full_word_count * MAX_BYTES
        data = np.frombuffer(buffer[buf_start:buf_start + buffer_size], dtype=np.uint32).copy()
        crypto.encrypt_blocks(data, self.key, full_word_count)
        buffer[buf_start:buf_start + buffer_size] = data.tobytes()

    def decrypt_full_blocks_parallel(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Convert buffer slice to numpy array for parallel processing
        buffer_size = full_word_count * MAX_BYTES
        data = np.frombuffer(buffer[buf_start:buf_start + buffer_size], dtype=np.uint32).copy()
        crypto.decrypt_blocks(data, self.key, full_word_count)
        buffer[buf_start:buf_start + buffer_size] = data.tobytes()

    @staticmethod
    def encrypt_words(n: int, key: np.ndarray, data: bytearray, offset: int) -> None:
        v = np.frombuffer(data[offset:offset + n*4], dtype=np.uint32).copy()
        crypto.encrypt_block(v, key, n)
        data[offset:offset + n*4] = v.tobytes()

    @staticmethod
    def decrypt_words(n: int, key: np.ndarray, data: bytearray, offset: int) -> None:
        v = np.frombuffer(data[offset:offset + n*4], dtype=np.uint32).copy()
        crypto.decrypt_block(v, key, n)
        data[offset:offset + n*4] = v.tobytes()

    @staticmethod
    def encrypt_fixed_word_array(key: np.ndarray, data: bytearray, offset: int) -> None:
        if len(data) - offset < MAX_BYTES:
            return

        v = np.frombuffer(data[offset:offset + MAX_BYTES], dtype=np.uint32).copy()
        crypto.encrypt_block_fixed(v, key)
        data[offset:offset + MAX_BYTES] = v.tobytes()

    @staticmethod
    def decrypt_fixed_word_array(key: np.ndarray, data: bytearray, offset: int) -> None:
        if len(data) - offset < MAX_BYTES:
            return

        v = np.frombuffer(data[offset:offset + MAX_BYTES], dtype=np.uint32).copy()
        crypto.decrypt_block_fixed(v, key)
        data[offset:offset + MAX_BYTES] = v.tobytes()




from typing import Dict, List, Iterable, Optional, BinaryIO

from .keys import KeyType, Mapping as KeyMapping
from .xxtea_writer import XXTEAWriter
from .xxtea_reader import XXTEAReader
from .constants import KNOWN_PLAIN
from .metadata import MetadataType
from .file import File
from .xtea import XTEA
from .utils import *

import zipfile
import secrets
import struct
import os
import io

class Osz2Package:
    def __init__(
        self,
        reader: Optional[BinaryIO] = None,
        metadata_only: bool = False,
        key_type: KeyType = KeyType.OSZ2
    ) -> None:
        self.metadata: Dict[MetadataType, str] = {}
        self.beatmap_ids: Dict[str, int] = {}
        self.files: List[File] = []
        self.key_type = key_type
        self.key: bytes = b""
        self.version: int = 0

        self.data_offset: int = 0
        self.file_info_offset: int = 0

        self.iv: bytes = secrets.token_bytes(16)
        self.metadata_hash: bytes = b""
        self.file_info_hash: bytes = b""
        self.full_body_hash: bytes = b""

        if reader is not None:
            self.read(reader, metadata_only)

    @classmethod
    def from_file(cls, path: str, metadata_only=False, key_type=KeyType.OSZ2) -> "Osz2Package":
        """Read an osz2 package from a file path"""
        with open(path, "rb") as f:
            return cls(f, metadata_only, key_type)

    @classmethod
    def from_bytes(cls, data: bytes, metadata_only=False, key_type=KeyType.OSZ2) -> "Osz2Package":
        """Read an osz2 package from raw bytes"""
        with io.BytesIO(data) as f:
            return cls(f, metadata_only, key_type)

    @classmethod
    def from_directory(
        cls,
        directory: str,
        key_type: KeyType = KeyType.OSZ2
    ) -> "Osz2Package":
        """Initialize an osz2 package object from a directory, useful for exporting packages"""
        package = cls(key_type=key_type)

        for root, _, filenames in os.walk(directory):
            for filename in filenames:
                filepath = os.path.join(root, filename)
                rel_path = os.path.relpath(filepath, directory)

                with open(filepath, 'rb') as f:
                    content = f.read()

                stat = os.stat(filepath)
                created = datetime.datetime.fromtimestamp(stat.st_ctime)
                modified = datetime.datetime.fromtimestamp(stat.st_mtime)

                file = File(
                    filename=rel_path,
                    offset=0,
                    size=len(content),
                    hash=hashlib.md5(content).digest(),
                    date_created=created,
                    date_modified=modified,
                    content=content
                )
                package.files.append(file)

                # Auto-assign beatmap IDs
                if file.is_beatmap:
                    package.beatmap_ids[rel_path] = -1

        return package

    @property
    def video_files(self) -> Iterable[File]:
        for file in self.files:
            if file.is_video:
                yield file

    @property
    def beatmap_files(self) -> Iterable[File]:
        for file in self.files:
            if file.is_beatmap:
                yield file

    @property
    def combined_beatmap_files(self) -> Iterable[File]:
        for file in self.files:
            if file.is_combined_beatmap:
                yield file

    @property
    def osz_filename(self, extension: str = ".osz") -> str:
        return sanitize_filename(
            f'{self.metadata.get(MetadataType.BeatmapSetID, "")} '
            f'{self.metadata.get(MetadataType.Artist, "Unknown")} - '
            f'{self.metadata.get(MetadataType.Title, "Unknown")} '
            f'({self.metadata.get(MetadataType.Creator, "Unknown")})'
        ).strip() + extension

    def read(self, reader: BinaryIO, metadata_only: bool = False) -> None:
        """Read osz2 package data from a reader & apply it to this object"""
        self._read_header(reader)

        if metadata_only:
            return

        self._read_files(reader)

    def export(self) -> bytes:
        """Export the current package as an osz2 file"""
        # Ensure we're ready for export
        assert len(self.files) > 0, "Cannot create an empty package"

        if self.key_type == KeyType.OSZ2:
            assert MetadataType.Creator in self.metadata, "Missing required metadata: 'Creator'"
            assert MetadataType.BeatmapSetID in self.metadata, "Missing required metadata: 'BeatmapSetID'"

        elif self.key_type == KeyType.OSF2:
            assert MetadataType.Title in self.metadata, "Missing required metadata: 'Title'"
            assert MetadataType.Artist in self.metadata, "Missing required metadata: 'Artist'"

        # Generate encryption key
        key_generator = KeyMapping[self.key_type]
        key = key_generator(self.metadata)
        key_array = bytes_to_uint32_array(key)
        self.key = key

        output = io.BytesIO()
        self._process_video_files()
        self._write_package_contents(output, key_array)
        return output.getvalue()

    def create_osz_package(
        self,
        compression: int = zipfile.ZIP_DEFLATED,
        exclude_disallowed_files: bool = True
    ) -> bytes:
        """Create a regular .osz package from the current files"""
        with io.BytesIO() as buffer:
            osz = zipfile.ZipFile(buffer, 'w', compression)

            for file in self.files:
                if exclude_disallowed_files and not file.is_allowed_extension:
                    # See `constants.ALLOWED_FILE_EXTENSIONS` for allowed file extensions
                    continue

                # Create ZipInfo to set file metadata
                zip_info = zipfile.ZipInfo(filename=file.filename)
                zip_info.compress_type = compression
                zip_info.date_time = file.date_modified.timetuple()[:6]

                # Check if date_time is valid for zip format
                if zip_info.date_time[0] < 1980:
                    zip_info.date_time = (1980, 1, 1, 0, 0, 0)

                osz.writestr(zip_info, file.content)

            osz.close()
            return buffer.getvalue()

    def calculate_osz_filesize(
        self,
        compression: int = zipfile.ZIP_DEFLATED,
        exclude_disallowed_files: bool = True
    ) -> int:
        """Calculate the size of the .osz package if it were to be created"""
        return len(self.create_osz_package(compression, exclude_disallowed_files))

    def find_file_by_name(self, name: str) -> Optional[File]:
        """Get a file by its filename"""
        return next((file for file in self.files if file.filename == name), None)

    def find_file_by_beatmap_id(self, beatmap_id: int) -> Optional[File]:
        """Get a file by its beatmap ID"""
        return next((file for file in self.files if self.beatmap_ids.get(file.filename, -1) == beatmap_id), None)

    def _read_header(self, reader: BinaryIO) -> None:
        magic = reader.read(3)
        assert magic == b"\xECHO", "Not a valid osz2 package" # nice one echo

        self.version = struct.unpack("B", reader.read(1))[0]
        self.iv = reader.read(16)

        self.metadata_hash = reader.read(16)
        self.file_info_hash = reader.read(16)
        self.full_body_hash = reader.read(16)

        self._read_metadata(reader)
        self._read_file_names(reader)

        # Generate key based on metadata and key type
        # Usually this is just the MD5 of "<creator>yhxyfjo5<beatmapsetID>"
        key_generator = KeyMapping[self.key_type]
        self.key = key_generator(self.metadata)

    def _read_metadata(self, reader: BinaryIO) -> None:
        buffer = reader.read(4)
        count = struct.unpack("<I", buffer)[0]

        for _ in range(count):
            buf = reader.read(2)
            meta_type = struct.unpack("<H", buf)[0]
            meta_value = read_string(reader)

            self.metadata[MetadataType(meta_type)] = meta_value

            buffer += buf
            buffer += write_string(meta_value)

        metadata_hash = compute_osz_hash(buffer, count*3, 0xA7)
        assert metadata_hash == self.metadata_hash, f"Metadata hash mismatch, expected: {metadata_hash.hex()}, got: {self.metadata_hash.hex()}"

    def _read_file_names(self, reader: BinaryIO) -> None:
        buffer = reader.read(4)
        count = struct.unpack("<I", buffer)[0]

        for _ in range(count):
            filename = read_string(reader)
            beatmap_id = struct.unpack("<I", reader.read(4))[0]
            self.beatmap_ids[filename] = beatmap_id

    def _read_files(self, reader: BinaryIO) -> None:
        # Convert key to uint32 array for XXTEA
        key = bytes_to_uint32_array(self.key)

        # Verify encrypted magic
        encrypted_magic = bytearray(reader.read(64))
        xtea = XTEA(key)
        xtea.decrypt(encrypted_magic, 0, 64)
        assert encrypted_magic == KNOWN_PLAIN, "Invalid encryption key"

        # Store file info offset
        self.file_info_offset = reader.tell()

        # Read encrypted i32 length
        length = struct.unpack("<I", reader.read(4))[0]

        # Decode length by encrypted length
        for i in range(0, 16, 2):
            length -= self.file_info_hash[i] | (self.file_info_hash[i+1] << 17)

        file_info = reader.read(length)

        # Store data offset
        self.data_offset = reader.tell()
        
        file_data = reader.read()
        file_offset = reader.seek(0, 1)
        total_size = reader.seek(0, 2)
        reader.seek(file_offset, 0)

        # Parse file infos using xxtea stream
        with XXTEAReader(io.BytesIO(file_info), key) as xxtea:
            count = struct.unpack("<I", xxtea.read(4))[0]
            curr_offset = struct.unpack("<I", xxtea.read(4))[0]

            # Verify file info hash
            file_info_hash = compute_osz_hash(file_info, count*4, 0xd1)
            assert file_info_hash == self.file_info_hash, f"File info hash mismatch, expected: {file_info_hash.hex()}, got: {self.file_info_hash.hex()}"

            for i in range(count):
                filename = read_string(xxtea)
                file_hash = xxtea.read(16)

                date_created_binary = struct.unpack("<Q", xxtea.read(8))[0]
                date_modified_binary = struct.unpack("<Q", xxtea.read(8))[0]

                # Convert from .NET DateTime.ToBinary() format
                date_created = datetime_from_binary(date_created_binary)
                date_modified = datetime_from_binary(date_modified_binary)

                next_offset = total_size - file_offset
                if count > i + 1:
                    next_offset = struct.unpack("<I", xxtea.read(4))[0]

                file_length = next_offset - curr_offset
                file = File(
                    filename,
                    curr_offset,
                    file_length,
                    file_hash,
                    date_created,
                    date_modified,
                    content=bytes(),
                )
                self.files.append(file)
                curr_offset = next_offset

        # After reading the file info, read the actual file contents
        with XXTEAReader(io.BytesIO(file_data), key) as xxtea:
            for i in range(len(self.files)):
                length = struct.unpack("<I", xxtea.read(4))[0]
                self.files[i].content = xxtea.read(length)

    def _write_package_contents(self, writer: BinaryIO, key: List[int]) -> None:
        # Prepare file data & info
        file_data = self._write_file_data(self.files, key)
        file_info = self._write_file_info(self.files, key)

        # Calculate hashes
        hash_info = compute_osz_hash(
            file_info,
            len(self.files) * 4, 0xD1
        )
        hash_body = compute_body_hash(
            file_data,
            int(self.metadata.get(MetadataType.VideoDataOffset, -1)) if MetadataType.VideoDataOffset in self.metadata else None,
            int(self.metadata.get(MetadataType.VideoDataLength, -1)) if MetadataType.VideoDataLength in self.metadata else None
        )

        # Prepare metadata
        meta_data = self._write_metadata()
        hash_meta = compute_osz_hash(meta_data, len(self.metadata) * 3, 0xA7)

        # Create & encode IV by XORing with body hash
        encoded_iv = bytes(a ^ b for a, b in zip(self.iv, hash_body))

        writer.write(b'\xECHO')
        writer.write(struct.pack("B", self.version))
        writer.write(encoded_iv)
        writer.write(hash_meta)
        writer.write(hash_info)
        writer.write(hash_body)
        writer.write(meta_data)

        beatmap_files = {
            f.filename: self.beatmap_ids.get(f.filename, -1) 
            for f in self.files if f.is_beatmap
        }
        writer.write(struct.pack("<I", len(beatmap_files)))

        for filename, beatmap_id in beatmap_files.items():
            writer.write(write_string(filename))
            writer.write(struct.pack("<I", beatmap_id & 0xFFFFFFFF))

        known_plain = bytearray(KNOWN_PLAIN)
        xtea = XTEA(key)
        xtea.encrypt(known_plain, 0, 64)
        writer.write(bytes(known_plain))

        # Obfuscate file_info length
        encoded_length = len(file_info)
        for i in range(0, 16, 2):
            encoded_length += hash_info[i] | (hash_info[i + 1] << 17)

        # File info & data with obfuscated length
        writer.write(struct.pack("<I", encoded_length & 0xFFFFFFFF))
        writer.write(file_info)
        writer.write(file_data)

    def _write_file_data(self, files: List[File], key: List[int]) -> bytes:
        with XXTEAWriter(key) as writer:
            for file in files:
                writer.write(struct.pack("<I", len(file.content)))
                writer.write(file.content)
            return writer.getvalue()

    def _write_file_info(self, files: List[File], key: List[int]) -> bytes:
        with XXTEAWriter(key) as writer:
            # Write file count
            writer.write(struct.pack("<I", len(files)))

            # Calculate all offsets first
            offsets = []
            offset = 0

            for file in files:
                offsets.append(offset)
                offset += 4 + len(file.content)

            # Write first offset (always 0)
            writer.write(struct.pack("<I", offsets[0]))

            for i, file in enumerate(files):
                # Write filename & hash
                writer.write_string(file.filename)
                writer.write(file.hash)

                # Write timestamps
                writer.write(struct.pack("<Q", datetime_to_binary(file.date_created)))
                writer.write(struct.pack("<Q", datetime_to_binary(file.date_modified)))

                # Write next offset (except for last file)
                if i < len(files) - 1:
                    writer.write(struct.pack("<I", offsets[i + 1]))

            return writer.getvalue()

    def _write_metadata(self) -> bytes:
        buffer = io.BytesIO()
        buffer.write(struct.pack("<I", len(self.metadata)))

        for meta_type, value in self.metadata.items():
            buffer.write(struct.pack("<H", int(meta_type)))
            buffer.write(write_string(value or ""))

        return buffer.getvalue()

    def _process_video_files(self) -> None:
        offset = 0

        for file in self.files:
            if not file.is_video:
                offset += 4 + len(file.content)
                continue

            assert len(file.content) >= 1024, "Video needs to be at least 1024 bytes big"

            # Calculate video hash from middle section of the file
            data_length = len(file.content)
            foot_start = (data_length // 2) - ((data_length // 2) % 16) - 512 + 16
            foot_data = file.content[foot_start:foot_start + 1024]
            video_hash = hashlib.md5(foot_data).hexdigest().upper()

            self.metadata[MetadataType.VideoDataOffset] = str(offset)
            self.metadata[MetadataType.VideoDataLength] = str(data_length)
            self.metadata[MetadataType.VideoHash] = video_hash
            break




from osz2.utils import bytes_to_uint32_array
from osz2.constants import KNOWN_PLAIN
from pathlib import Path
from osz2 import *

import secrets
import pytest
import os

TESTS_DIR = Path(__file__).parent
FILES_DIR = TESTS_DIR / "files"
OSZ2_FILES = list(FILES_DIR.glob("*.osz2"))

class TestSimpleCryptor:
    @pytest.fixture
    def key(self) -> list:
        return bytes_to_uint32_array(secrets.token_bytes(16))

    def test_encrypt_decrypt_roundtrip(self, key: list) -> None:
        cryptor = SimpleCryptor(key)
        original = bytearray(b"plz enjoy game")
        data = bytearray(original)

        cryptor.encrypt_bytes(data)
        assert data != original

        cryptor.decrypt_bytes(data)
        assert data == original

    def test_encrypt_modifies_data(self, key: list) -> None:
        cryptor = SimpleCryptor(key)
        data = bytearray(b"test data")
        original = bytearray(data)

        cryptor.encrypt_bytes(data)
        assert data != original

    def test_empty_data(self, key: list) -> None:
        cryptor = SimpleCryptor(key)
        data = bytearray()
        cryptor.encrypt_bytes(data)
        cryptor.decrypt_bytes(data)
        assert len(data) == 0

class TestXXTEA:
    @pytest.fixture
    def key(self) -> list:
        return bytes_to_uint32_array(secrets.token_bytes(16))

    def test_encrypt_decrypt_roundtrip(self, key: list) -> None:
        xxtea = XXTEA(key)
        original = bytearray(KNOWN_PLAIN)
        data = bytearray(original)

        xxtea.encrypt(data, 0, len(data))
        assert data != original

        xxtea.decrypt(data, 0, len(data))
        assert data == original

    def test_encrypt_decrypt_large_data(self, key: list) -> None:
        xxtea = XXTEA(key)
        original = bytearray(os.urandom(1024))
        data = bytearray(original)

        xxtea.encrypt(data, 0, len(data))
        assert data != original

        xxtea.decrypt(data, 0, len(data))
        assert data == original

    def test_encrypt_partial_buffer(self, key: list) -> None:
        xxtea = XXTEA(key)
        original = bytearray(b"\x00" * 10 + os.urandom(64) + b"\x00" * 10)
        data = bytearray(original)

        # Encrypt only middle portion
        xxtea.encrypt(data, 10, 64)

        # Start and end should be unchanged
        assert data[:10] == original[:10]
        assert data[74:] == original[74:]

        # Middle should be different
        assert data[10:74] != original[10:74]

class TestXTEA:
    @pytest.fixture
    def key(self) -> list:
        return bytes_to_uint32_array(secrets.token_bytes(16))

    def test_encrypt_decrypt_roundtrip(self, key: list) -> None:
        xtea = XTEA(key)
        original = bytearray(os.urandom(64))
        data = bytearray(original)

        xtea.encrypt(data, 0, len(data))
        assert data != original

        xtea.decrypt(data, 0, len(data))
        assert data == original




from .xxtea import XXTEA
from io import BytesIO
from typing import (
    BinaryIO,
    Iterable,
    Optional,
    Iterator,
    List
)

class XXTEAReader(BinaryIO):
    """XXTEA decryption reader that decrypts data in chunks"""

    def __init__(self, reader: BytesIO, key: List[int]) -> None:
        self.reader: BytesIO = reader
        self.xxtea: XXTEA = XXTEA(key)

    def __enter__(self) -> "XXTEAReader":
        return self

    def __exit__(self, exc_type: object, exc_value: object, traceback: object) -> None:
        self.reader.close()

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    @property
    def closed(self) -> bool:
        return self.reader.closed

    @property
    def mode(self) -> str:
        return "rb"

    @property
    def name(self) -> str:
        return getattr(self.reader, 'name', '<xxtea_stream>')

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return self.reader.seekable()

    def read(self, n: int = -1) -> bytes:
        read = bytearray(self.reader.read(n))
        if len(read) > 0:
            self.xxtea.decrypt(read, 0, len(read))
        return bytes(read)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.reader.seek(offset, whence)

    def tell(self) -> int:
        return self.reader.tell()

    def flush(self) -> None:
        self.reader.flush()

    def close(self) -> None:
        self.reader.close()

    def fileno(self) -> int:
        return self.reader.fileno()

    def isatty(self) -> bool:
        return self.reader.isatty()

    def readline(self, limit: int = -1) -> bytes:
        raise NotImplementedError("readline is not supported for encrypted streams")

    def readlines(self, hint: int = -1) -> List[bytes]:
        raise NotImplementedError("readlines is not supported for encrypted streams")

    def write(self, s: bytes) -> int:  # type: ignore[override]
        raise NotImplementedError("XXTEAReader is read-only")

    def writelines(self, lines: Iterable[bytes]) -> None:  # type: ignore[override]
        raise NotImplementedError("XXTEAReader is read-only")

    def truncate(self, size: Optional[int] = None) -> int:
        raise NotImplementedError("XXTEAReader is read-only")



"""Type stubs for the crypto C extension module."""

import numpy.typing as npt
import numpy as np

def encrypt_bytes(buf: bytearray, key: npt.NDArray[np.uint8]) -> None: ...
def decrypt_bytes(buf: bytearray, key: npt.NDArray[np.uint8]) -> None: ...
def encrypt_blocks(data: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], full_word_count: int) -> None: ...
def decrypt_blocks(data: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], full_word_count: int) -> None: ...
def encrypt_block(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], n: int) -> None: ...
def decrypt_block(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], n: int) -> None: ...
def encrypt_block_fixed(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32]) -> None: ...
def decrypt_block_fixed(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32]) -> None: ...




from typing import Tuple, Dict, Union, Optional
from osz2.metadata import MetadataType
from osz2.package import Osz2Package
from osz2.keys import KeyType
from pathlib import Path

import argparse
import typing
import sys
import os

def main() -> None:
    parser = argparse.ArgumentParser(prog="osz2", description="A tool to decrypt, extract, and create osz2 files")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    decrypt_parser = subparsers.add_parser("decrypt", help="Decrypt and extract an osz2 file")
    decrypt_parser.add_argument("input", help="The path to the osz2 file to decrypt")
    decrypt_parser.add_argument("output", help="The path to put the extracted files")
    decrypt_parser.add_argument("--key-type", choices=["osz2", "osf2"], default="osz2", help="The key generation method")
    decrypt_parser.add_argument("--create-osz", action="store_true", help="Also create a regular .osz package")

    encrypt_parser = subparsers.add_parser("encrypt", help="Create an osz2 package from a directory")
    encrypt_parser.add_argument("input", help="The path to the directory containing files")
    encrypt_parser.add_argument("output", help="The output path for the osz2 file")
    encrypt_parser.add_argument("--key-type", choices=["osz2", "osf2"], default="osz2", help="The key generation method")

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        return

    key_type = KeyType(args.key_type)

    if args.command == "decrypt":
        osz2 = decrypt_osz2(args.input, key_type)
        save_osz2(osz2, args.output)

        if not args.create_osz:
            return

        with open(f"{args.output}/{osz2.osz_filename}", "wb") as f:
            osz_data = osz2.create_osz_package(exclude_disallowed_files=False)
            f.write(osz_data)

    elif args.command == "encrypt":
        encrypt_directory(
            args.input,
            args.output,
            key_type
        )

def decrypt_osz2(filepath: str, key_type: KeyType) -> Osz2Package:
    if not os.path.exists(filepath):
        print(f"Error: Input file does not exist: {filepath}", file=sys.stderr)
        sys.exit(1)

    print("Reading osz2 package...")
    return Osz2Package.from_file(filepath, key_type=key_type)

def save_osz2(package: Osz2Package, output: str) -> None:
    Path(output).mkdir(exist_ok=True)
    print(f"Extracting {len(package.files)} files to {output}")

    for file in package.files:
        output_path = os.path.join(output, file.filename)

        if (directory := Path(output_path).parent) != ".":
            directory.mkdir(parents=True, exist_ok=True)

        with open(output_path, "wb") as f:
            f.write(file.content)

        print(f"  -> {file.filename} ({len(file.content)} bytes)")

def encrypt_directory(
    directory: str,
    output: str,
    key_type: KeyType
) -> None:
    if not os.path.exists(directory):
        print(f"Error: Input directory does not exist: {directory}", file=sys.stderr)
        sys.exit(1)

    print(f"Creating osz2 package from directory: {directory}")
    package = Osz2Package.from_directory(directory, key_type)

    # Try to parse beatmap metadata and apply it to package
    for beatmap in package.beatmap_files:
        _, content = parse_beatmap(beatmap.content.decode("utf-8-sig"))
        beatmap_id = content.get('Metadata', {}).get('BeatmapID', None)

        if beatmap_id is not None:
            package.beatmap_ids[beatmap.filename] = int(beatmap_id)

        apply_metadata(package, content)

    print(f"Exporting package with {len(package.files)} files...")

    with open(output, "wb") as f:
        data = package.export()
        f.write(data)

    print(f"Saved to: {output}")

def apply_metadata(package: Osz2Package, beatmap: Dict[str, dict]) -> None:
    if 'Metadata' not in beatmap:
        print("Error: No 'Metadata' section found in beatmap")
        sys.exit(1)

    if 'General' not in beatmap:
        print("Error: No 'General' section found in beatmap")
        sys.exit(1)

    metadata_section = beatmap['Metadata']
    title = metadata_section.get('TitleUnicode') or metadata_section.get('Title', '')
    artist = metadata_section.get('ArtistUnicode') or metadata_section.get('Artist', '')
    creator = metadata_section.get('Creator', '')
    source = metadata_section.get('Source', '')
    tags = metadata_section.get('Tags', '')
    beatmapset_id = metadata_section.get('BeatmapSetID', -1)

    general_section = beatmap['General']
    preview_time = general_section.get('PreviewTime', 0)

    package.metadata[MetadataType.Title] = str(title)
    package.metadata[MetadataType.Artist] = str(artist)
    package.metadata[MetadataType.Creator] = str(creator)
    package.metadata[MetadataType.Source] = str(source)
    package.metadata[MetadataType.Tags] = str(tags)
    package.metadata[MetadataType.BeatmapSetID] = str(beatmapset_id)
    package.metadata[MetadataType.PreviewTime] = str(preview_time)

@typing.no_type_check
def parse_beatmap(content: str) -> Tuple[int, Dict[str, dict]]:
    sections: Dict[str, Union[dict, list]] = {}
    current_section = None
    beatmap_version = 0

    for line in content.splitlines():
        if line.startswith('osu file format'):
            beatmap_version = int(line.removeprefix('osu file format v'))
            continue

        if line.startswith('[') and line.endswith(']'):
            # New section
            current_section = line.removeprefix('[').removesuffix(']')
            continue

        if current_section is None:
            continue

        if not line:
            continue

        if current_section in ('General', 'Editor', 'Metadata', 'Difficulty'):
            if current_section not in sections:
                sections[current_section] = {}

            # Parse key, value pair
            key, raw_value = (
                split.strip() for split in line.split(':', maxsplit=1)
            )

            # Try to parse float/int
            parsed_value = parse_number(raw_value) or raw_value
            section_dict: dict = sections[current_section]
            section_dict[key] = parsed_value
            continue

        if current_section not in sections:
            sections[current_section] = []

        # Append to list
        section_list: list = sections[current_section]
        section_list.append(line)

    return beatmap_version, sections

def parse_number(value: str) -> Optional[Union[int, float]]:
    for cast in (int, float):
        try:
            return cast(value.strip())
        except ValueError:
            continue
    return None

if __name__ == "__main__":
    main()




from .utils import write_uleb128
from .xxtea import XXTEA
from io import BytesIO
from typing import (
    BinaryIO,
    Iterable,
    Optional,
    Iterator,
    List
)

class XXTEAWriter(BinaryIO):
    """XXTEA encryption writer that encrypts data in chunks"""

    def __init__(self, key: List[int]) -> None:
        self.buffer: BytesIO = BytesIO()
        self.xxtea: XXTEA = XXTEA(key)

    def __enter__(self) -> "XXTEAWriter":
        return self

    def __exit__(self, exc_type: object, exc_value: object, traceback: object) -> None:
        self.buffer.close()

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    @property
    def closed(self) -> bool:
        return self.buffer.closed

    @property
    def mode(self) -> str:
        return "wb"

    @property
    def name(self) -> str:
        return getattr(self.buffer, 'name', '<xxtea_stream>')

    def readable(self) -> bool:
        return False

    def writable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return self.buffer.seekable()

    def write(self, data: bytes) -> int:  # type: ignore[override]
        if len(data) <= 0:
            return 0

        buf = bytearray(data)
        self.xxtea.encrypt(buf, 0, len(buf))
        return self.buffer.write(bytes(buf))

    def write_string(self, s: str) -> None:
        encoded = s.encode('utf-8')
        length = len(encoded)
        self.write(write_uleb128(length))
        self.write(encoded)

    def writelines(self, lines: Iterable[bytes]) -> None:  # type: ignore[override]
        for line in lines:
            self.write(line)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.buffer.seek(offset, whence)

    def tell(self) -> int:
        return self.buffer.tell()

    def flush(self) -> None:
        self.buffer.flush()

    def close(self) -> None:
        self.buffer.close()

    def fileno(self) -> int:
        return self.buffer.fileno()

    def isatty(self) -> bool:
        return self.buffer.isatty()

    def read(self, n: int = -1) -> bytes:
        raise NotImplementedError("XXTEAWriter is write-only")

    def readline(self, limit: int = -1) -> bytes:
        raise NotImplementedError("XXTEAWriter is write-only")

    def readlines(self, hint: int = -1) -> List[bytes]:
        raise NotImplementedError("XXTEAWriter is write-only")

    def truncate(self, size: Optional[int] = None) -> int:
        return self.buffer.truncate(size)

    def getvalue(self) -> bytes:
        return self.buffer.getvalue()



from setuptools import Extension, setup  # type: ignore[import-untyped]

setup(
    ext_modules=[
        Extension(
            name="osz2.crypto",
            sources=["osz2/crypto.c"],
        )
    ]
)




from osz2 import Osz2Package, MetadataType
from pytest import fixture
from pathlib import Path

TESTS_DIR = Path(__file__).parent
FILES_DIR = TESTS_DIR / "files"
OSZ2_FILES = list(FILES_DIR.glob("*.osz2"))

class TestOsz2Package:
    @fixture(params=OSZ2_FILES, ids=lambda p: p.stem)
    def osz2_path(self, request) -> Path:
        return request.param

    def test_from_file(self, osz2_path: Path) -> None:
        package = Osz2Package.from_file(str(osz2_path))
        assert package is not None
        assert len(package.files) > 0
        assert len(package.metadata) > 0

    def test_from_bytes(self, osz2_path: Path) -> None:
        data = osz2_path.read_bytes()
        package = Osz2Package.from_bytes(data)
        assert package is not None
        assert len(package.files) > 0

    def test_metadata_only(self, osz2_path: Path) -> None:
        package = Osz2Package.from_file(str(osz2_path), metadata_only=True)
        assert len(package.metadata) > 0
        assert len(package.files) == 0

    def test_metadata_contains_required_fields(self, osz2_path: Path) -> None:
        package = Osz2Package.from_file(str(osz2_path), metadata_only=True)
        # 'Creator' and 'BeatmapSetID' are required for osz2 key generation
        assert MetadataType.Creator in package.metadata
        assert MetadataType.BeatmapSetID in package.metadata

    def test_create_osz_package(self, osz2_path: Path) -> None:
        package = Osz2Package.from_file(str(osz2_path))
        osz_data = package.create_osz_package()
        assert len(osz_data) > 0
        assert osz_data[:2] == b"PK"

    def test_beatmap_content_is_valid(self, osz2_path: Path) -> None:
        package = Osz2Package.from_file(str(osz2_path))

        for beatmap in package.beatmap_files:
            # osu! beatmaps should start with "osu file format"
            content = beatmap.content.decode("utf-8-sig", errors="replace")
            assert "osu file format" in content

class TestOsz2ExportRoundtrip:
    @fixture(params=OSZ2_FILES, ids=lambda p: p.stem)
    def osz2_path(self, request) -> Path:
        return request.param

    def test_export_and_reimport(self, osz2_path: Path) -> None:
        original = Osz2Package.from_file(str(osz2_path))
        exported = original.export()
        reimported = Osz2Package.from_bytes(exported)

        # Compare metadata & file count
        assert reimported.metadata == original.metadata
        assert len(reimported.files) == len(original.files)

        # Compare file contents
        for original_file in original.files:
            reimported_file = reimported.find_file_by_name(original_file.filename)
            assert reimported_file is not None
            assert reimported_file.content == original_file.content
